# $Id: __init__.py,v 1.8 2019/06/25 06:51:04 ypguo Exp $

"""libraries to manipulate the devices and network."""

__author__ = 'Yongping Guo <guoyoooping@163.com>'
__copyright__ = 'Copyright (c) 2017 Yongping Guo'
__license__ = 'GPLv3'
__url__ = 'https://github.com/guoyoooping/netdevice'
__version__ = '0.13'

__all__ = ['linux', 'junos', 'cisco'] 

#from netdevice import *
#import junos
#import linux
#import cisco
#import webpage
